;;; Proposals for other files
;;; Dmitriy Ivanov: divanov_nosp@m_aha.ru (replace _nosp@m_ by @)

(in-package :pdf)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; font-metrics.lisp
;;; Make extract-font-metrics-encoding generic?

(defun extract-font-metrics-encoding (font-metrics)
  (let ((encoding (or (get-encoding (encoding-scheme font-metrics))
		      (get-encoding (font-name font-metrics)))))
    (if encoding
	encoding
	(make-instance 'single-byte-encoding :name (font-name font-metrics)
		       :standard-encoding nil
		       :char-names (map 'vector #'(lambda (char)
						    (and char (name char)))
					(encoding-vector font-metrics))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; font.lisp
;;; Customize your lisp implementation in treating charset.
;;; The get-char-metrics generic is intended to replace get-char.

(defgeneric get-char-metrics (char-or-code font encoding)
 ;;; Args: char-or-code  Lisp character or its char-code
 (:method (char-or-code font encoding)
  (aref (characters font) (force-char-code char-or-code))) )

(defmethod get-char-metrics ((code integer) font (encoding custom-encoding))
  (aref (characters font)
        (let ((charset (charset encoding)))
          (if charset
              #+lispworks (ef:char-external-code (code-char code) charset)
              #-lispworks code
              code))))

(defmethod get-char-metrics ((char character) font (encoding custom-encoding))
 ;;; Map Unicode char code to code belonging [0-255] range.
  (aref (characters font)
        (let ((charset (charset encoding)))
          (if charset
              #+lispworks (ef:char-external-code char charset)
              #-lispworks (char-code char)
              (char-code char)))))

(defun get-char-width (char-or-code font &optional font-size)
  (let ((char-metrics (get-char-metrics char-or-code font (encoding font))))
    (if font-size (* (width char-metrics) font-size) (width char-metrics))))

(defun get-char-size (char-or-code font &optional font-size)
  (let* ((char-metrics (get-char-metrics char-or-code font (encoding font)))
	 (width (width char-metrics))
	 (bbox (bbox char-metrics))
	 (ascender (aref bbox 3))
	 (descender (aref bbox 1)))
    (if font-size
	(values (* width font-size)(* ascender font-size)(* descender font-size))
	(values width ascender descender))))

(defun get-char-italic-correction (char-or-code font &optional font-size)
  (let* ((char-metrics (get-char-metrics char-or-code font (encoding font)))
	 (left (left-italic-correction char-metrics))
	 (right (right-italic-correction char-metrics)))
    (if font-size
	(values (* left font-size)(* right font-size))
	(values left right))))

(defun get-kerning (char1 char2 font &optional font-size)
  (let* ((encoding (encoding font))
         (char-metrics1 (get-char-metrics char1 font encoding))
         (char-metrics2 (get-char-metrics char2 font encoding))
         (kerning (gethash (+ (ash (code char-metrics1) 16) (code char-metrics2))
                           (kernings font)
                           0)))
    (if font-size (* font-size kerning) kerning)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; pdf.lisp

(defun pdf-string (obj &key (unicode :default))
 ;;; Used to embrace a pdf string used in places other than content streams,
  ;; e.g. titles, annotations etc.
  ;; Args: unicode  If true or defaults to true, the PDF text string is encoded in Unicode
  ;;	   lang ?
  ;; Q: Rename or create separate pdf-text-string?
  (let ((string (if (stringp obj) obj (princ-to-string obj))))
    (when (eq unicode :default)
      (setq unicode (notevery #+lispworks #'lw:base-char-p
                              #-lispworks (lambda (char) (<= (char-code char) 255))
                              string)))
    (with-output-to-string (stream nil :element-type 'base-char)
      (write-char #\( stream)
      (when unicode			; write the Unicode byte order marker U+FEFF
        (write-char #.(code-char 254) stream) (write-char #.(code-char 255) stream))
      (loop for char across string
            for code = (char-code char)
            if unicode
            do (write-char (code-char (ldb (byte 8 8) code)) stream)	; hi
               (write-char (code-char (ldb (byte 8 0) code)) stream)	; lo
            else if (> code 255)
            do (write-char (code-char (ldb (byte 8 0) code)) stream)	; lo
            else do (case char ((#\( #\) #\\)
                                (write-char #\\ stream)))
                      (write-char char stream))
      (write-char #\) stream))))


(defmethod write-stream-content ((content string))
  ;; Args: content Base string, may include
  ;;	   - either one-byte codes (already converted to external format if needed)
  ;;	   - or Unicode two-byte character codes (big-endian CIDs)
  #+pdf-binary
  (loop for char across content
        do (write-byte (ldb (byte 8 0) (char-code char)) *pdf-stream*))
  #-pdf-binary
  (write-sequence obj *pdf-stream*))

(defmethod write-document ((filename pathname) &optional (document *document*))
   (with-open-file (stream filename
                           :direction :output :if-exists :supersede
                           :element-type #+pdf-binary '(unsigned-byte 8) 
                                         #-pdf-binary 'base-char
                           :external-format +external-format+)
     (write-document stream document)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; pdf-base.lisp

;;; String output

(defgeneric write-to-page (object encoding &optional escape)
 (:documentation
  "Write object (only text for now) into current *page-stream* in text mode"))

;;; show-text writes string as is, draw-text escapes dangerous characters.
;;; Args: text String or a single character

(defun show-text (text)
  (if *font*
      (progn (write-to-page text (if *font* (encoding *font*)))
        (write-line "Tj" *page-stream*))
      (format *page-stream* "(~a) Tj~%" text)))

(defun draw-text (text)
  (if *font*
      (progn (write-to-page text (if *font* (encoding *font*)) t)
        (write-line "Tj" *page-stream*))
      (format *page-stream* "(~a) Tj~%" text)))

(defun show-spaced-strings (strings)
  (write-string "[ " *page-stream*)
  (let ((encoding (if *font* (encoding *font*))))
    (dolist (item strings)
      (if (numberp item)
          (format *page-stream* "~a " item)
          (write-to-page item encoding))))
  (write-line "] TJ" *page-stream*))

(defun show-text-on-next-line (string)
  (write-to-page string (if *font* (encoding *font*)))
  (write-line "'" *page-stream*))


(defmethod write-to-page :before ((string string) encoding &optional escape)
  (declare (ignore escape))
  (write-char #\( *page-stream*))

(defmethod write-to-page :after ((string string) encoding &optional escape)
  (declare (ignore escape))
  (write-string ") " *page-stream*))

(defmethod write-to-page ((string string) encoding &optional escape)
  (if escape
      (loop for char across string
            do (case char
                 ((#\( #\) #\\)
                  (write-char #\\ *page-stream*)
                  (write-char char *page-stream*))
                 ;(#\Newline
                 ; (write-string "\\n" *page-stream*))
                 ;(#\Return
                 ; (write-string "\\r" *page-stream*))
                 ;(#\Tab
                 ; (write-string "\\t" *page-stream*))
                 (otherwise
                  (write-char char *page-stream*))))
      (write-string string *page-stream*)))

(defmethod write-to-page ((string string) (encoding custom-encoding) &optional escape)
  (if (or escape #+lispworks (lw:text-string-p string))	; may include unicode
      (loop with charset = (charset encoding)
            for char across string do
            (case char
              ((#\( #\) #\\)
               (when escape
                 (write-char #\\ *page-stream*))
               (write-char char *page-stream*))
              (otherwise
               (write-char (if (and charset #+lispworks (not (lw:base-char-p char)))
                               #+lispworks (code-char(ef:char-external-code char charset))
                               #-lispworks char
                               char)			; write-byte would be great
                           *page-stream*))))
      (write-string string *page-stream*)))


(defmethod write-to-page ((string string) (encoding unicode-encoding) &optional escape)
  (declare (ignore escape))
  (loop for char across string
        for code = (char-code char)
        do (write-char (code-char (ldb (byte 8 8) code)) *page-stream*)
           (write-char (code-char (ldb (byte 8 0) code)) *page-stream*)))

;;; Single character output

(defun show-char (char)
 ;;; Deprecated in favor of show-text or draw-text
  (write-to-page char (if *font* (encoding *font*)) t)
  (write-line "Tj" *page-stream*))

(defmethod write-to-page :before ((char character) encoding &optional escape)
  (declare (ignore escape))
  (write-char #\( *page-stream*))

(defmethod write-to-page :after ((char character) encoding &optional escape)
  (declare (ignore escape))
  (write-char #\) *page-stream*))

(defmethod write-to-page ((char character) encoding &optional escape)
  (when escape (case char
                 ((#\( #\) #\\) (write-char #\\ *page-stream*))))
  (write-char char *page-stream*))

(defmethod write-to-page ((char character) (encoding custom-encoding) &optional escape)
  (when escape (case char
                 ((#\( #\) #\\) (write-char #\\ *page-stream*))))
  (write-char (let ((charset (charset encoding)))
                (if (and charset #+lispworks (not (lw:base-char-p char)))
                    #+lispworks (code-char (ef:char-external-code char charset))
                    #-lispworks char
                    char))
              *page-stream*))

(defmethod write-to-page ((char character) (encoding unicode-encoding) &optional escape)
  (when escape (case char
                 ((#\( #\) #\\) ;(write-char #\Nnull *page-stream*) ?
                  (write-char #\\ *page-stream*))))
  (let ((code (char-code char)))
    (write-char (code-char (ldb (byte 8 8) code)) *page-stream*)
    (write-char (code-char (ldb (byte 8 0) code)) *page-stream*)))

